% Constrained Support Vector Clustering:
% Use the clusters generated from a mixture of factor analyzers (MFA)
% to provide better constraints (or kernel matrix) for the one class
% classification (extended to support vector clustering)
function ConstrainedSVC()
    close all;
    clear all;

    file = 'data_unifang_radialgauss1.data';
    %file = 'dataset7.data';
    %file = 'dataset1.data';
    [X, Y] = loadDataXY(file);
    plotOrigData(X,Y);

    % run the mixture of factor analyzers and get the 
    mixComponents = 20;  % number of gaussian components 
    lowerDimension = 2;  % lower dimension k of the factor analyzers
    Params = getMfaParams(X, mixComponents,lowerDimension);
    plotMfa(X, Params);
    
    % run support vector clustering constraining with the parameters
    % obtained by mfa
    nu = 0.08; gamma = 10; clusters = 2;   
    %nu = 0.08; gamma = 9.4; %best params!!   
    runSVC(X, Params, nu, gamma, Y, clusters);
    
function runSVC(X, Params, nu, gamma, Y, clusters)
    % compute kernel matrix, measuring the distance to the mfa cluster
    % centers
    Lbls = Params{1};
    clustMeans = Params{2};
    %K = kernel
    [X1, I] = reguralizeData(X, Params);
    Y1 = Y(I, :);
    plotOrigData(X1,Y1);

    %X_ = X1;
    X_ = X;
    % run the constrained support vector clustering on the cleaned data
    % todo: change! this should be run as regularized problem
    SvmParams = runLibSvm(X_, nu, gamma);
    plotClustering(X_, SvmParams{1}, SvmParams{2}, SvmParams{3});
    Lbls = connectClusters(X_, SvmParams, clusters);
    figure;
    hold on;
    plotClustering(X_, SvmParams{1}, SvmParams{2}, Lbls);
    
function Lbls = connectClusters(X, Params, clusters)
    Lbls = Params{3};
    uniqueLbls = unique(Lbls);
    num_lbls = size(uniqueLbls,1);
    if (find(uniqueLbls == 0))
        num_lbls = num_lbls - 1;
        uniqueLbls = uniqueLbls(2:size(uniqueLbls,1));
    end;
    if (num_lbls <= clusters)
        return;
    end;
    % merge two clusters at a time until there are just two clusters
    % overall
    while (num_lbls > clusters)
        [c1, c2] = closestClusters(X, Lbls);
        inds = find(Lbls == c2);
        Lbls(inds) = c1;
        num_lbls = num_lbls -1;
    end;
    L = Lbls(find(Lbls ~= 0));
    minLbl = min(L);
    maxLbl = max(L);
    indmin = find(Lbls == minLbl);
    indmax = find(Lbls == maxLbl);
    Lbls(indmin,1) = 1;
    Lbls(indmax,1) = 2;
    
function [c1, c2] = closestClusters(X, Lbls)
    unique_lbls = unique(Lbls);
    num_lbls = size(unique_lbls);
    if (find(Lbls == 0)) % there is a bounded support vector
        start = 2;
    else
        start = 1;
    end;
    c1 = 1; c2 = 1;
    mindist = realmax;
    for i = start:num_lbls-1
        inds1 = find(Lbls == unique_lbls(i));        
        for j = i + 1:num_lbls
            inds2 = find(Lbls == unique_lbls(j));
            tmpdist = minPairwiseDist(X(inds1,:), X(inds2,:));
            if (tmpdist < mindist)
                c1 = unique_lbls(i);
                c2 = unique_lbls(j);
                mindist = tmpdist;
            end;
        end;
    end;
    
function D = minPairwiseDist(X, Y)    
    D = realmax;
    for i = 1:size(X,1)
        rX = repmat(X(i,:), size(Y,1),1);
        dists = (rX-Y).*(rX-Y);
        dists = sum(dists,2);
        tmpdist = min(dists);
        if (tmpdist < D)
            D = tmpdist;
        end;
    end;
    
% returns the the datapoints that are close to their cluster center (the 
% center is computed from the MFA assignment) and their indices    
function [X1, I] = reguralizeData(X, Params)
    Lbls = Params{1};
    ClusterMeans = Params{2};
    CovMatrices = Params{3};
    lbls_type = unique(Lbls);
    num_lbls = size(lbls_type,1);
    X1 = [];
    I = [];
    for i = 1:num_lbls
        inds = find(Lbls(:,1) == lbls_type(i));
        meanCluster = ClusterMeans(lbls_type(i),:);
        examp = X(inds,:);
        if (size(examp,1) < 2)
            continue;
        end;
        covMatr = cov(examp);
        stdX = std(examp); 
        for j = 1:size(examp,1)
            pt = examp(j,:);
            ptDev = abs(pt-mean(examp))
            LessThanStd = 1;
            for p = 1:size(ptDev,2)
                if ptDev(1,p) > 1.5*stdX(1,p)
                    LessThanStd = 0;
                    break;
                end;
            end;
            if (LessThanStd==1)
                X1 = [X1;pt];
                I = [I; inds(j,1)];
            end;
        end;
    end;

function Params = getMfaParams(X, mixComponents, lowerDimension)    
    [Lh,Ph,Mu,Pi,LL]=mfa(X,mixComponents,lowerDimension);
    Params = clusterDataMfa(X, Lh, Ph, Mu, Pi); % returns a cell array. The first element the labels, the second - the cluster means, the third - the covariances
    
function Ind = getIndices(X, svs)
    Ind = [];
    for i = 1:size(svs,1)
        tmpInd = findInd(X, svs(i,:));
        Ind = [Ind; tmpInd];
    end;
    
function Ind = findInd(X, sv)
    Ind = [];
    for i = 1:size(X,1)
        isInd = 1;
        for j = 1:size(X,2)
            if (X(i,j) ~= sv(1,j))
                isInd = 0;
                break;
            end;
        end;
        if (isInd == 1)
            Ind = i;
            return;
        end;
    end;
    
function [X, Y] = loadDataXY(file)
    dims_data = 2;
    dims_file = 3;
    D = loadData(file,dims_file);
    cols= size(D,2);
    X = D(:,1:cols-1);
    Y = D(:,cols);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Plotting functions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function plotMfa(X, Params)
    figure;
    Lbls = Params{1};
    ClusterMeans = Params{2};
    CovMatrices = Params{3};
    lbls_type = unique(Lbls);
    num_lbls = size(lbls_type,1);
    color_map = ['b'; 'r'; 'g'; 'm'; 'y'; 'k'; 'c']; 
    hold on;
    for i = 1:num_lbls
        inds = find(Lbls(:,1) == lbls_type(i));
        col = rem(i-1, size(color_map,1))+1;
        col = color_map(col,1); 
        plot(X(inds,1),X(inds,2), ['.' col], 'LineWidth', 1);
        meanCluster = ClusterMeans(lbls_type(i),:);
        examp = X(inds,:);
        covMatr = cov(examp);
        size(examp)
        if size(examp,1) < 2
            continue;
        end;
        draw_ellipse(meanCluster, covMatr, col, 1);  
        hold on;
    end;
    axis([-4,4,-4,4]); 
    
function plotOrigData(X, Y)    
    inds = find(Y == 1);
    figure;
    plot(X(inds,1), X(inds,2), '.b');
    inds = find(Y==2);
    hold on;
    plot(X(inds,1), X(inds, 2), '.r');   
    title('Test data');
    axis([-4,4,-4,4]);    
    
function plotClustering(X, model, train_label, Lbls)
    svs = full(model.SVs);
    sv_inds = getIndices(X, svs);
    inds_neg = find(train_label < 0);
    inds_pos = find(train_label > 1);
    %figure;
    plot(svs(:,1), svs(:,2), 'ok'); 
    lbls_type = unique(Lbls);
    num_lbls = size(lbls_type,1);
    color_map = ['b'; 'r'; 'g'; 'm'; 'y'; 'c']; 
    for i = 1:num_lbls
        inds = find(Lbls(:,1) == lbls_type(i,1));
        if (lbls_type(i,1) == 0) % plot the BSVs
            scol = '.k';
        else 
            col = rem(i-1, size(color_map,1))+1;
            scol = ['.' color_map(col,1)];
        end;
        plot(X(inds,1),X(inds,2), scol, 'LineWidth', 1);        
    end;
    %axis([-4,4,-4,4]);    
    
function plotContours(X, model, gamma)
    minx = min(X(:,1));
    maxx = max(X(:,1));
    miny = min(X(:,2));
    maxy = max(X(:,2));
    stepx = (maxx - minx)/50;
    stepy = (maxy - miny)/50;
    Xcoord = minx:stepx:maxx;
    Ycoord = miny:stepy:maxy;
    Z = zeros(size(Ycoord,2), size(Xcoord,2));
    for i = 1:size(Ycoord,2)
        for j = 1:size(Xcoord,2)
            Z(i,j) = model.sv_coef'*KernelG([Xcoord(1,j) Ycoord(1,i)], model, gamma) - model.rho;    
            if (Z(i,j) < -0.05)
                Z(i,j) = -0.05;
            elseif (Z(i,j) > 1)
                Z(i,j) = 1;
            end;
        end;
    end;
    figure;
    [Xm,Ym] = meshgrid(Xcoord,Ycoord);
    [c, h]=contour3(Xm,Ym,Z,[0 0.1 0.2 0.3]);
    set(h, 'LineWidth', 2)
    hold on;
    surface(Xm,Ym,Z,'EdgeColor',[.8 .8 .8],'FaceColor','none')        
    contour3(Xm,Ym,Z,[0 0.1 0.2 0.3]);    
    colormap cool    
    grid off
    view(35,65)
    axis tight
    colormap cool

    figure;
    %Z = Z*100;
    surfcc(Xm,Ym,Z)
    colormap cool
    view(55,65)
    hold on;
    
    figure;
    [c, h] = contour(Xm, Ym, Z, [0 0]);
    set(h, 'LineWidth', 2);
    hold on;
    plot(X(:,1),X(:,2), '.k', 'LineWidth', 1);        
    k = 1;
    
    